/*
    libcentroidal_profile provides functions to calculate centroidal profiles
    of regions and match them against one another.
    Copyright (C) 2004  Sebastian Hans <hanss@in.tum.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __H_CENTROIDAL_PROFILE__
#define __H_CENTROIDAL_PROFILE 23

#include <opencv/cv.h>

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************
 * Types
 ******************************************************************************/
typedef struct _cp_profile
{
	long num_points;     /* Number of points in the profile */
	double *data;        /* Profile data consisting of num_points double
				values, dynamically allocated, is freed on
				destruction of the profile */
	double avg_distance; /* Average distance of points on the circumference
				from the the center */
	double orientation;  /* Orientation of the image (in radians) */
}
cp_profile_t;

typedef struct _cp_model
{
	long id;                /* Some arbitrary ID value */
	char *name;             /* A string identifying the model, can be NULL,
				   is freed on request */
	void *user_data;	/* Some arbitrary user data, can be NULL, is
				   freed on request */
	IplImage *image;        /* Original model image, 8-bit greyscale, is
				   released on destruction of the model */
	cp_profile_t *profile;  /* Profile of the model, is destroyed on
				   destruction of the model */
	struct _cp_model *prev; /* Pointer to the previous model in the model
				   list or NULL if this is the first model */
	struct _cp_model *next; /* Pointer to the next model in the model list
				   or NULL if this is the last model */
}
cp_model_t;

typedef struct _cp_model_list
{
	long length;       /* Number of elements in the model list */
	long num_points;   /* Number of points for the profiles of the models in
			      the list */
	cp_model_t *first; /* First model in the list or NULL if length == 0 */
	cp_model_t *last;  /* Last model in the list or NULL if length == 0 */
}
cp_model_list_t;

typedef struct _cp_match
{
	cp_model_t *model;           /* Matched model, is *not* freed on
					destruction of the match */
	double difference;           /* Measure of the difference between the
					matched image and the model */
	double relative_orientation; /* Orientation of the image relative to the
					orientation of the model (in radians) */
	double relative_size;        /* Size of the image relative to the size
					of the model */
	struct _cp_match *prev;      /* Pointer to the previous match in the
					match list or NULL if this is the first
					match */
	struct _cp_match *next;      /* Pointer to the next match in the match
					list or NULL if this is the last match
				      */
}
cp_match_t;

typedef struct _cp_match_list
{
	long length;       /* Number of elements in the match list */
	cp_match_t *first; /* First match in the list or NULL if length == 0 */
	cp_match_t *last;  /* Last match in the list or NULL if length == 0 */
}
cp_match_list_t;

/******************************************************************************
 * match_list.c
 ******************************************************************************/

/*
 * cp_find_match_for_model returns the match in match_list corresponding to the
 * model model or NULL if such a match can not be found.
 * Also returns NULL if either match_list or model are NULL.
 */
extern cp_match_t *cp_find_match_for_model(cp_match_list_t *match_list,
		cp_model_t *model);

/*
 * cp_find_match_for_model_id returns the match in match_list corresponding to
 * the model with the ID id or NULL if such a match can not be found.
 * Also returns NULL if match_list is NULL.
 */
extern cp_match_t *cp_find_match_for_model_id(cp_match_list_t *match_list,
		long id);

/*
 * cp_find_match_for_model_name returns the match in match_list corresponding
 * to the model with the name name or NULL if such a match can not be found.
 * Also returns NULL if match_list is NULL.
 */
extern cp_match_t *cp_find_match_for_model_name(cp_match_list_t *match_list,
		char *name);

/*
 * cp_sort_match_list sorts a match list by difference (ascending).
 * Currently, it uses bubble sort.
 */
extern void cp_sort_match_list(cp_match_list_t *match_list);

/*
 * cp_destroy_match_list frees the match list and all matches contained therein.
 */
extern void cp_destroy_match_list(cp_match_list_t *match_list);

/*******************************************************************************
 * model.c
 ******************************************************************************/


/*
 * cp_create_model creates a model like cp_create_model_with_user_data with
 * user_data set to NULL.
 */
extern cp_model_t *cp_create_model(IplImage *image, long id, char *name,
		long num_points);

/*
 * cp_create_model_with_user_data creates a model with num_points points from
 * image. id is an arbitrary model ID. name is a string or NULL. user_data is a
 * pointer to arbitrary user data.
 * image will be taken over by the model and released when the model is
 * destroyed.
 *
 * The image is interpreted as a binary image, i.e. every pixel with a non-zero
 * value is considered part of the model. Pixels with value zero are background.
 * The image must consist of exactly one conntected region of non-zero pixels
 * for this to work.
 *
 * May return NULL if memory could not be allocated or profile generation
 * failed.
 */
extern cp_model_t *cp_create_model_with_user_data(IplImage *image, long id,
		char *name, long num_points, void *user_data);

/*
 * cp_set_num_points_for_model recalculates the profile for model with
 * num_points points. Returns 1 on success, 0 on failure. In case of failure,
 * the profile is not modified. If num_points equals the current number of
 * points of the profile, the profile is *not* recalculated.
 */
extern int cp_set_num_points_for_model(cp_model_t *model, long num_points);

/*
 * cp_destroy_model frees the model and its image as well as its profile.
 * If free_name has a non-zero value, the name attribute of the model will also
 * be freed. If free_user_data has a non-zero value, the user_data attribute of
 * the model will also be freed.
 */
extern void cp_destroy_model(cp_model_t *model, int free_name,
		int free_user_data);

/*******************************************************************************
 * model_list.c
 ******************************************************************************/

/*
 * cp_create_model_list creates an empty model list and returns a pointer to it.
 * The models in the model list will all be created with num_points points.
 * The created model list should be destroyed using cd_destroy_model_list.
 *
 * May return NULL if memory could not be allocated.
 */
extern cp_model_list_t *cp_create_model_list(long num_points);

/*
 * cp_set_num_points_for_model_list tries to recalculate the profile for every
 * model in the list with num_points points.
 * Returns 1 on success, 0 on failure. In case of failure, all models for which
 * the recalculation failed are dropped from the list and destroyed.
 *
 * Use with care!
 *
 * If num_points equals the current number of points for the list, nothing is
 * done.
 */
extern int cp_set_num_points_for_model_list(cp_model_list_t *model_list,
		long num_points);

/*
 * cp_append_model is short for cp_append_model_with_user_data with user_data
 * set to NULL.
 */
extern int cp_append_model(cp_model_list_t *model_list, IplImage *image,
		long id, char *name);
/*
 * cp_append_model_with_user_data creates a model from image, which must be an
 * 8-bit greyscale image, and appends it to model_list. The model will have the
 * number of points specified when model_list was created. id is an arbitrary
 * model ID. name can is a string or NULL. user_data is a pointer to arbitrary
 * user data. image will be taken over by the model and released when the model
 * is destroyed.
 *
 * Returns 1 on success, 0 otherwise. In case of failure, model_list is not
 * modified.
 */
extern int cp_append_model_with_user_data(cp_model_list_t *model_list,
		IplImage *image, long id, char *name, void *user_data);

/*
 * cp_find_model_id returns the first model in model_list that has the ID id or
 * NULL if such a model can not be found.
 * Also returns NULL if model_list is NULL.
 */
extern cp_model_t *cp_find_model_id(cp_model_list_t *model_list, long id);

/*
 * cp_find_model_name returns the first model in model_list that has the name
 * name or NULL if such a model can not be found.
 * Also returns NULL if model_list is NULL.
 */
extern cp_model_t *cp_find_model_name(cp_model_list_t *model_list, char *name);

/*
 * cp_remove_model removes the model from model_list. model must be in the list.
 * This is not checked, so be careful! The removed model is returned. It should
 * be destroyed with cp_destroy_model.
 * May return NULL if either model_list or model is NULL.
 */
extern cp_model_t *cp_remove_model(cp_model_list_t *model_list,
		cp_model_t *model);

/*
 * cp_destroy_model_list frees the model list and all models contained therein.
 * The images contained in the models as well as the profiles are also freed.
 * If free_name has a non-zero value, the name attributes of all the models will
 * also be freed. If free_user_data has a non-zero value, the user_data
 * attributes of all the models will also be freed.
 */
extern void cp_destroy_model_list(cp_model_list_t *model_list, int free_name,
		int free_user_data);

/*******************************************************************************
 * profile.c
 ******************************************************************************/

/*
 * cp_create_profile creates a profile fro the image image, which must be an
 * 8-bit greyscale image. It will be interpreted as a binary image with one
 * connected region. The profile will have num_points points.
 *
 * The image is interpreted as a binary image, i.e. every pixel with a non-zero
 * value is considered part of the model. Pixels with value zero are background.
 * The image must consist of exactly one conntected region of non-zero pixels
 * for this to work.
 *
 * May return NULL if memory could not be allocated.
 */
extern cp_profile_t *cp_create_profile(IplImage *image, long num_points);

/*
 * cp_recalculate_profile recalculates the profile for image and num_points
 * points. Returns 1 on success, 0 on failure. In case of failure, the profile
 * is not modified. If num_points equals the current number of points of the
 * profile, the profile is *not* recalculated.
 */
extern int cp_recalculate_profile(cp_profile_t *profile, IplImage *image,
		long num_points);

/*
 * cp_destroy_profile frees the profile along with its data.
 */
extern void cp_destroy_profile(cp_profile_t *profile);

/*
 * cp_match matches an image against a list of models. The image must be an
 * 8-bit greyscale image. It will be interpreted as a binary image with one
 * connected region (see cp_create_profile). cp_match creates a match list which
 * contains one match for every model in model_list and returns a pointer to it.
 * The match list should be destroyed with cp_destroy_match_list.
 * If best_match is not NULL, it will be assigned the best match in the returned
 * list.
 *
 * May return NULL if either image or model_list is NULL, model_list is empty,
 * or profile generation for the image failed.
 */
extern cp_match_list_t *cp_match(IplImage *image, cp_model_list_t *model_list,
		cp_match_t **best_match);

#ifdef __cplusplus
}
#endif

#endif
