/*
    Example program for libcentroidal_profile.
    Copyright (C) 2004  Sebastian Hans <hanss@in.tum.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <centroidal_profile.h>
#include <opencv/highgui.h>     /* cvLoadImage */
#include <stdio.h>              /* fprintf, printf, stderr */
#include <math.h>		/* M_PI_2 */

cp_model_list_t *model_list;

int load_models(int argc, char **argv)
{
	int i;
	IplImage *image;

	/* Create model list for 23 point models */
	if (!(model_list = cp_create_model_list(23)))
		return 0;

	for (i = 0; i < argc; i++)
	{
		/* Load model image. This is a function of OpenCV. */
		if (!(image = cvLoadImage(argv[i], 0)))
		{
			fprintf(stderr, "Couldn't load model image '%s'.\n",
					argv[i]);
			cp_destroy_model_list(model_list, 0, 0);
			return 0;
		}

		/* Create model and append it to the list. */
		if (!cp_append_model(model_list, image, i, argv[i]))
		{
			fprintf(stderr, "Couldn't append model '%s'.\n",
					argv[i]);
			cp_destroy_model_list(model_list, 0, 0);
			return 0;
		}
	}
	return 1;
}

int main(int argc, char **argv)
{
	IplImage *image;
	cp_match_list_t *match_list;
	cp_match_t *best_match, *match;
	char *name_of_best_model;

	if (argc < 3)
	{
		fprintf(stderr, "Usage: %s image modelimage [modelimage ...]\n",
				argv[0]);
		fprintf(stderr, "  for 8-bit greyscale images.\n");
		return 1;
	}
	if (!load_models(argc - 2, argv + 2))
		return 1;

	/* Load image to match. This is a function of OpenCV. */
	if (!(image = cvLoadImage(argv[1], 0)))
	{
		fprintf(stderr, "Couldn't load image '%s'.\n", argv[1]);
		cp_destroy_model_list(model_list, 0, 0);
		return 1;
	}

	/*
	 * match_list is a list containing one match for each of the models.
	 * We are only interested in best_match, which is the element of
	 * match_list with the least difference.
	 */
	if (!(match_list = cp_match(image, model_list, &best_match)))
	{
		fprintf(stderr, "Match failed. Out of memory?\n");
		cp_destroy_model_list(model_list, 0, 0);
		return 1;
	}

	printf("All matches:\n");
	for (match = match_list->first; match; match = match->next)
		printf("\t%-30s: diff = %g\n", match->model->name,
				match->difference);

	/* Sort matches by difference. */
	cp_sort_match_list(match_list);

	printf("\nMatches sorted by difference:\n");
	for (match = match_list->first; match; match = match->next)
		printf("\t%-30s: diff = %g\n", match->model->name,
				match->difference);

	printf("\nBest match was '%s':\n\tdiff = %g, orient = %g, size = %g\n",
			best_match->model->name, best_match->difference,
			best_match->relative_orientation * 90 / M_PI_2,
			best_match->relative_size);


	/* Free lists. */
	cp_destroy_match_list(match_list);
	cp_destroy_model_list(model_list, 0, 0);

	return 0;
}
