/*
    libcentroidal_profile provides functions to calculate centroidal profiles
    of regions and match them against one another.
    Copyright (C) 2004  Sebastian Hans <hanss@in.tum.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <centroidal_profile_internal.h>

#include <stdlib.h> /* malloc, free */
#include <string.h> /* strcmp */

cp_match_list_t *cp_create_match_list(void)
{
	cp_match_list_t *match_list;

	if (!(match_list = malloc(sizeof(cp_match_list_t))))
		return NULL;

	match_list->length = 0;
	match_list->first = match_list->last = NULL;

	return match_list;
}

void cp_append_match(cp_match_list_t *match_list, cp_match_t *match)
{
	if (!match_list || !match)
		return;

	match_list->length++;
	match->prev = match->next = NULL;

	if (match_list->length == 1)
		match_list->first = match_list->last = match;
	else
	{
		match_list->last->next = match;
		match->prev = match_list->last;
		match_list->last = match;
	}
}

cp_match_t *cp_find_match_for_model(cp_match_list_t *match_list,
		cp_model_t *model)
{
	cp_match_t *match;

	if (!match_list || !match_list->length)
		return NULL;

	for (match = match_list->first; match; match = match->next)
		if (match->model == model)
			return match;

	return NULL;
}

cp_match_t *cp_find_match_for_model_id(cp_match_list_t *match_list, long id)
{
	cp_match_t *match;

	if (!match_list || !match_list->length)
		return NULL;

	for (match = match_list->first; match; match = match->next)
		if (match->model->id == id)
			return match;

	return NULL;
}

cp_match_t *cp_find_match_for_model_name(cp_match_list_t *match_list,
		char *name)
{
	cp_match_t *match;

	if (!match_list || !match_list->length)
		return NULL;

	for (match = match_list->first; match; match = match->next)
		if (!name)
		{
			if (!match->model->name)
				return match;
		}
		else if (match->model->name &&
				!strcmp(match->model->name, name))
			return match;

	return NULL;
}

void cp_sort_match_list(cp_match_list_t *match_list)
{
	long i;
	cp_match_t *match, *next;

	if (!match_list || match_list->length < 2)
		return;

	for (i = 1; i < match_list->length; i++)
	{
		match = match_list->first;
		while (match && match->next)
		{
			if (match->difference > match->next->difference)
			{
				next = match->next;
				next->prev = match->prev;
				if (match->prev)
					match->prev->next = next;
				match->next = next->next;
				if (next->next)
					next->next->prev = match;
				match->prev = next;
				next->next = match;
				if (match == match_list->first)
					match_list->first = next;
				if (next == match_list->last)
					match_list->last = match;
			}
			else
				match = match->next;
		}
	}
}

static void cp_destroy_match_recursively(cp_match_t *match)
{
	cp_match_t *next;

	while (match)
	{
		next = match->next;
		cp_destroy_match(match);
		match = next;
	}
}

void cp_destroy_match_list(cp_match_list_t *match_list)
{
	if (!match_list)
		return;
	if (match_list->length)
		cp_destroy_match_recursively(match_list->first);
	free(match_list);
}
