/*
    libcentroidal_profile provides functions to calculate centroidal profiles
    of regions and match them against one another.
    Copyright (C) 2004  Sebastian Hans <hanss@in.tum.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <centroidal_profile.h>

#include <stdlib.h> /* malloc, free */
#include <string.h> /* strcmp */

cp_model_list_t *cp_create_model_list(long num_points)
{
	cp_model_list_t *model_list;

	if (!(model_list = malloc(sizeof(cp_model_list_t))))
		return NULL;

	model_list->length = 0;
	model_list->num_points = num_points;
	model_list->first = model_list->last = NULL;

	return model_list;
}

int cp_set_num_points_for_model_list(cp_model_list_t *model_list,
		long num_points)
{
	cp_model_t *model;
	int deleted_one_in_this_round, deleted_one = 0;

	if (!model_list || num_points == model_list->num_points)
		return 1;
	model_list->num_points = num_points;
	do
	{
		deleted_one_in_this_round = 0;
		for (model = model_list->first; model; model = model->next)
			if (!cp_set_num_points_for_model(model, num_points))
			{
				cp_remove_model(model_list, model);
				deleted_one = 1;
				break;
			}
	}
	while (deleted_one_in_this_round);
	return !deleted_one;
}

int cp_append_model(cp_model_list_t *model_list, IplImage *image, long id,
		char *name)
{
	return cp_append_model_with_user_data(model_list, image, id, name,
			NULL);
}

int cp_append_model_with_user_data(cp_model_list_t *model_list, IplImage *image,
		long id, char *name, void *user_data)
{
	cp_model_t *model;

	if (!model_list || !image)
		return 0;
	if (!(model = cp_create_model_with_user_data(image, id, name,
					model_list->num_points, user_data)))
		return 0;

	model_list->length++;
	model->prev = model->next = NULL;

	if (model_list->length == 1)
		model_list->first = model_list->last = model;
	else
	{
		model_list->last->next = model;
		model->prev = model_list->last;
		model_list->last = model;
	}
	return 1;
}

cp_model_t *cp_find_model_id(cp_model_list_t *model_list, long id)
{
	cp_model_t *model;

	if (!model_list || !model_list->length)
		return NULL;

	for (model = model_list->first; model; model = model->next)
		if (model->id == id)
			return model;

	return NULL;
}

cp_model_t *cp_find_model_name(cp_model_list_t *model_list, char *name)
{
	cp_model_t *model;

	if (!model_list || !model_list->length)
		return NULL;

	for (model = model_list->first; model; model = model->next)
		if (!name)
		{
			if (!model->name)
				return model;
		}
		else if (model->name && !strcmp(model->name, name))
			return model;

	return NULL;
}

cp_model_t *cp_remove_model(cp_model_list_t *model_list, cp_model_t *model)
{
	if (!model_list || !model)
		return NULL;

	model_list->length--;

	if (model_list->first == model)
		model_list->first = model->next;
	if (model_list->last == model)
		model_list->last = model->prev;
	if (model->prev)
		model->prev->next = model->next;
	if (model->next)
		model->next->prev = model->prev;

	model->prev = model->next = NULL;

	return model;
}

static void cp_destroy_model_recursively(cp_model_t *model, int free_name,
		int free_user_data)
{
	cp_model_t *next;

	while (model)
	{
		next = model->next;
		cp_destroy_model(model, free_name, free_user_data);
		model = next;
	}
}

void cp_destroy_model_list(cp_model_list_t *model_list, int free_name,
		int free_user_data)
{
	if (!model_list)
		return;
	if (model_list->length)
		cp_destroy_model_recursively(model_list->first, free_name,
				free_user_data);
	free(model_list);
}
