/*
    libcentroidal_profile provides functions to calculate centroidal profiles
    of regions and match them against one another.
    Copyright (C) 2004  Sebastian Hans <hanss@in.tum.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <centroidal_profile_internal.h>

#include <stdlib.h> /* malloc, free */
#include <math.h>   /* atan2, fabs, hypot, M_PI, M_PI_2, M_SQRT2 */

static inline double cp_normalize_orientation(double orient)
{
	while (orient < -M_PI)
		orient += M_PI;

	while (orient > M_PI)
		orient -= M_PI;

	return orient;
}

static inline double cp_calculate_center_and_orientation(IplImage *image,
		double *x_center, double *y_center)
{
	CvMoments moments;

	if (!image)
		return 0.0;

	cvMoments(image, &moments, 1);

	if (x_center)
		*x_center = cvGetSpatialMoment(&moments, 1, 0) /
			cvGetSpatialMoment(&moments, 0, 0);
	if (y_center)
		*y_center = cvGetSpatialMoment(&moments, 0, 1) /
			cvGetSpatialMoment(&moments, 0, 0);

	return cp_normalize_orientation(-0.5 * atan2(2.0 *
				cvGetNormalizedCentralMoment(&moments, 1, 1),
				cvGetNormalizedCentralMoment(&moments, 0, 2) -
				cvGetNormalizedCentralMoment(&moments, 2, 0)));
}

static inline void cp_copy_contour(CvContour *contour, int *x, int *y)
{
	CvSeqReader reader;
	int point_count, i;

	cvStartReadSeq((CvSeq*) contour, &reader, 0);
	point_count = contour->total;

	for (i = 0; i < point_count; i++)
	{
		CvPoint p;

		p = CV_CURRENT_POINT(reader);
		CV_NEXT_SEQ_ELEM(sizeof(CvPoint), reader);
		x[i] = p.x;
		y[i] = p.y;
	}
}

static inline long cp_get_contour(IplImage *image, int **x, int **y)
{
	CvMemStorage *mem_storage;
	CvContour *contour = NULL;
	int point_count, i;

	cvFindContours(image, mem_storage = cvCreateMemStorage(0),
			(CvSeq**) &contour, sizeof(CvContour), CV_RETR_EXTERNAL,
			CV_CHAIN_APPROX_NONE);
	if (!contour)
		return 0;
	point_count = contour->total;
	if (!(*x = malloc(point_count * sizeof(int))))
		return 0;
	if (!(*y = malloc(point_count * sizeof(int))))
	{
		free(x);
		return 0;
	}
	cp_copy_contour(contour, *x, *y);
	cvReleaseMemStorage(&mem_storage);
	return point_count;
}

static inline long cp_calculate_starting_point(int *x, int *y, long length,
		double x_center, double y_center, double orientation)
{
	double best_orientation, best_distance, old_orientation;
	double current_orientation, current_distance;
	long best_index, i;

	best_index = length - 1;
	current_orientation = best_orientation =
		atan2(((double) x[length - 1]) - x_center,
				((double) y[length - 1]) - y_center);
	best_distance = -1.0;

#define UPDATE_BEST	do \
			{ \
				current_distance = hypot(((double) x[i]) - \
						x_center, ((double) y[i]) - \
						y_center); \
				if (current_distance > best_distance) \
				{ \
					best_index = i; \
					best_orientation = \
						current_orientation; \
					best_distance = current_distance; \
				} \
			} while (0)

	for (i = 0; i < length; i++)
	{
		old_orientation = current_orientation;
		current_orientation = atan2(((double) x[i]) - x_center,
				((double) y[i]) - y_center);

		if (old_orientation < orientation &&
					current_orientation >= orientation)
		{
			if (old_orientation < -M_PI_2 &&
					current_orientation > M_PI_2)
				; /* wrap-around */
			else
				UPDATE_BEST;
		}
		else if (old_orientation > orientation &&
				 current_orientation <= orientation)
		{
			if (old_orientation > M_PI_2 &&
					current_orientation < -M_PI_2)
				; /* wrap-around */
			else
				UPDATE_BEST;
		}
		else if (old_orientation < -M_PI_2 &&
				current_orientation > M_PI_2 &&
				old_orientation > orientation &&
				current_orientation > orientation)
			UPDATE_BEST;
		else if (old_orientation > M_PI_2 &&
				current_orientation < -M_PI_2 &&
				old_orientation < orientation &&
				current_orientation < orientation)
			UPDATE_BEST;
	}

	return best_index;
}

static inline double cp_calculate_circumference(int *x, int *y, long length)
{
	long i, straight_steps = 0, diagonal_steps = 0;

	for (i = 1; i < length; i++)
	{
		if (x[i] != x[i - 1] && y[i] != y[i - 1])
			diagonal_steps++;
		else
			straight_steps++;
	}

	if (x[i] != x[0] && y[i] != y[0])
		diagonal_steps++;
	else
		straight_steps++;

	return ((double) straight_steps) + ((double) diagonal_steps) * M_SQRT2;
}

static inline double cp_fill_profile_data(cp_profile_t *profile, int *x, int *y,
		long i_start, long length, double x_center, double y_center)
{
	long num_points, i, j, k;
	double distance_between_points, distance_to_last = 0.0;
	double max_distance = 0.0;
	double *data;

	data = profile->data;
	num_points = profile->num_points;
	data[0] = hypot(((double) x[0]) - x_center, ((double) y[0]) - y_center);
	distance_between_points = cp_calculate_circumference(x, y, length) /
		((double) profile->num_points);

	for (i = i_start + 1, j = 1, k = 1; j < length && k < num_points;
			j++, i = (i + 1) % length)
	{
		distance_to_last += (x[i] != x[i - 1] && y[i] != y[i - 1] ?
				M_SQRT2 : 1.0);
		if (distance_to_last >= distance_between_points)
		{
			double current_distance;

			current_distance = hypot(((double) x[i]) - x_center,
					((double) y[i]) - y_center);
			if (current_distance > max_distance)
				max_distance = current_distance;
			while (distance_to_last >= distance_between_points)
			{
				distance_to_last -= distance_between_points;
				data[k++] = current_distance;
			}
		}
	}

	return max_distance;
}

static inline double cp_normalize_profile(cp_profile_t *profile, double max)
{
	double *data;
	long i, n;
	double total_distance = 0.0;

	data = profile->data;
	n = profile->num_points;

	for (i = 0; i < n; i++)
	{
		total_distance += data[i];
		data[i] = data[i] / max;
	}

	return total_distance / ((double) n);
}

static inline int cp_calculate_profile(cp_profile_t *profile, IplImage *image)
{
	double x_center, y_center, distance_between_samples, max_distance;
	int *x, *y;
	long length, i_start;

	if (!profile || !image)
		return 0;

	profile->orientation = cp_calculate_center_and_orientation(image,
			&x_center, &y_center);
	if (!(length = cp_get_contour(image, &x, &y)))
		return 0;
	i_start = cp_calculate_starting_point(x, y, length, x_center, y_center,
			profile->orientation);
	max_distance = cp_fill_profile_data(profile, x, y, i_start, length,
			x_center, y_center);
	profile->avg_distance = cp_normalize_profile(profile, max_distance);

	return 1;
}

static int cp_make_profile_data(cp_profile_t *profile, IplImage *image)
{
	if (!profile || !image)
		return 0;
	if (!(profile->data = malloc(profile->num_points * sizeof(double))))
		return 0;
	if (!cp_calculate_profile(profile, image))
	{
		free(profile->data);
		return  0;
	}
	return 1;
}

cp_profile_t *cp_create_profile(IplImage *image, long num_points)
{
	cp_profile_t *profile;

	if (!image)
		return NULL;

	if (!(profile = malloc(sizeof(cp_profile_t))))
		return NULL;

	profile->num_points = num_points;
	profile->avg_distance = 0.0;

	if (!cp_make_profile_data(profile, image))
	{
		cp_destroy_profile(profile);
		return NULL;
	}

	return profile;
}

int cp_recalculate_profile(cp_profile_t *profile, IplImage *image,
		long num_points)
{
	double *old_data, old_orient, old_avg_dist;
	long old_num_points;

	if (!profile || !image)
		return 0;

	if (num_points == profile->num_points)
		return 1;

	old_data = profile->data;
	old_num_points = profile->num_points;
	old_orient = profile->orientation;
	old_avg_dist = profile->avg_distance;

	profile->num_points = num_points;
	if (!cp_make_profile_data(profile, image))
	{
		profile->data = old_data;
		profile->num_points = old_num_points;
		profile->orientation = old_orient;
		profile->avg_distance = old_avg_dist;
		return 0;
	}
	free(old_data);
	return 1;
}

static inline double cp_match_profile_data(double *data1, double *data2,
		long num_points)
{
	long i;
	double diff = 0.0, local_diff;

	for (i = 0; i < num_points; i++)
	{
		local_diff = data1[i] - data2[i];
		diff += local_diff * local_diff;
	}
	return diff;
}

static inline double cp_relative_orientation(double orient1, double orient2)
{
	return cp_normalize_orientation(orient1 - orient2);
}

static inline cp_match_t *cp_match_profile_against_model(cp_profile_t *profile,
		cp_model_t *model)
{
	cp_match_t *match;

	if (!profile || !model ||
			profile->num_points != model->profile->num_points)
		return NULL;
	if (!(match = cp_create_match(model)))
		return NULL;

	match->difference = cp_match_profile_data(profile->data,
			model->profile->data, profile->num_points);
	match->relative_orientation =
		cp_relative_orientation(profile->orientation,
				model->profile->orientation);
	match->relative_size = profile->avg_distance /
		model->profile->avg_distance;

	return match;
}

static inline cp_match_list_t *cp_match_profile_against_list(
		cp_profile_t *profile, cp_model_list_t *model_list,
		cp_match_t **best_match)
{
	cp_match_list_t *match_list;
	cp_model_t *model;
	cp_match_t *match;

	if (best_match)
		*best_match = NULL;
	if (!profile || !model_list || !model_list->length ||
			profile->num_points != model_list->num_points ||
			!(match_list = cp_create_match_list()))
		return NULL;
	for (model = model_list->first; model; model = model->next)
		if ((match = cp_match_profile_against_model(profile, model)))
		{
			cp_append_match(match_list, match);
			if (best_match && (!*best_match ||
						match->difference <
						(*best_match)->difference))
				*best_match = match;
		}
		else
		{
			cp_destroy_match_list(match_list);
			if (best_match)
				*best_match = NULL;
			return NULL;
		}
	return match_list;
}

cp_match_list_t *cp_match(IplImage *image, cp_model_list_t *model_list,
		cp_match_t **best_match)
{
	IplImage *cloned_image;
	cp_profile_t *image_profile;
	cp_match_list_t *match_list;

	if (!image || !model_list || !model_list->length ||
			!(cloned_image = cvCloneImage(image)))
		return NULL;
	if (!(image_profile = cp_create_profile(cloned_image,
					model_list->num_points)))
	{
		cvReleaseImage(&cloned_image);
		return NULL;
	}
	cvReleaseImage(&cloned_image);

	match_list = cp_match_profile_against_list(image_profile, model_list,
			best_match);
	cp_destroy_profile(image_profile);
	return match_list;
}

void cp_destroy_profile(cp_profile_t *profile)
{
	if (!profile)
		return;
	if (profile->data)
		free(profile->data);
	free(profile);
}
